import numpy as np
import cupy as cp

from cucounter_C import HashTable

class Counter(HashTable):
    def __init__(self, keys, capacity: int = 0, capacity_factor: float = 1.7331):
        assert isinstance(keys, (np.ndarray, cp.ndarray)), "keys must be numpy or cupy ndarray"
        assert keys.dtype == np.uint64, "keys.dtype must be uint64"
        assert capacity_factor > 1.0, "capacity_factor must be greater than 1.0"

        # Dynamically determine hashtable capacity if not provided
        if capacity == 0:
            capacity = int(keys.size * capacity_factor) 

        assert capacity > keys.size, "Capacity must be greater than size of keyset"

        if isinstance(keys, np.ndarray):
            super().__init__(keys, capacity)
        elif isinstance(keys, cp.ndarray):
            super().__init__(keys.data.ptr, keys.size, capacity)

    def count(self, keys, count_revcomps=False, kmer_size=32):
        """
        Count a set of keys. 
        Duplicate keys are counted equally as many times as they are observed. 
        If count_revcomps == True, each individual key and its reverse complement will be counted
        as two separate keys. When computing the reverse complement of a key, the kmer_size
        argument determines the size of the kmer.

        Args:
            keys (numpy.ndarray or cupy.ndarray): the keys to increment in the hashtable.
            count_revcomps (bool): whether to count each key's reverse complement.
            kmer_size (int): the assumed kmer size.
        """
        assert isinstance(keys, (np.ndarray, cp.ndarray)), "keys must be numpy or cupy ndarray"
        assert keys.dtype == np.uint64, "keys.dtype must be uint64"
        assert kmer_size > 0 and kmer_size <= 32, "kmer size must be 32 >= size > 0"

        if isinstance(keys, np.ndarray):
            super().count(keys, count_revcomps, kmer_size)
        elif isinstance(keys, cp.ndarray):
            super().count(keys.data.ptr, keys.size, count_revcomps, kmer_size)

    def __getitem__(self, keys):
        """
        Look up a set of keys to get their corresponding counts from the hashtable.
        Each input key's index will correspond to the output count's indices.

        Args:
            keys (numpy.ndarray or cupy.ndarray): the keyset to lookup in the hashtable.
        Returns:
            counts (numpy.ndarray or cupy.ndarray): the counts corresponding to the input keyset.
                counts will match the input key's type when determining whether to return a
                numpy.ndarray or cupy.ndarray.
        """
        assert isinstance(keys, (np.ndarray, cp.ndarray)), "keys must be numpy or cupy ndarray"
        assert keys.dtype == np.uint64, "keys.dtype must be uint64"

        if isinstance(keys, np.ndarray):
            return super().get(keys)
        elif isinstance(keys, cp.ndarray):
            counts = cp.zeros_like(keys, dtype=np.uint32)
            super().get(keys.data.ptr, counts.data.ptr, keys.size)
            return counts 

    def get_probe_lengths(self, keys):
        """
        Determines the number of (linear probing) probes needed before termination for each key 
        in the input keyset. Termination for a key is reached when the probing scheme finds the
        corresponding key in the hashtable or when an empty position is found, indicating that the
        key is not present in the hashtable.
        Currently does not support checking probing lengths for reverse complements.

        Args:
            keys (numpy.ndarray or cupy.ndarray): the keyset to check probing lengths for.
        Returns:
            probe_lengths (numpy.ndarray or cupy.ndarray): the number of probes performed before
                terminating for each key. 
                probe_lengths will match the input key's type when determining whether to 
                return a numpy.ndarray or cupy.ndarray.
        """
        assert isinstance(keys, (np.ndarray, cp.ndarray)), "keys must be numpy or cupy ndarray"
        assert keys.dtype == np.uint64, "keys.dtype must be uint64"

        if isinstance(keys, np.ndarray):
            return super().get_probe_lengths(keys)
        elif isinstance(keys, cp.ndarray):
            probe_lengths = cp.zeros_like(keys, dtype=np.uint32)
            super.get_probe_lengths(keys.data.ptr, probe_lengths.data.ptr, keys.size)
            return probe_lengths

    def __repr__(self):
        return super().to_string(True)

    def __str__(self):
        return super().to_string(False)



import os
import sys
import re
import platform
import subprocess

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext


CUDA_ERROR_CHECK = True
USE_COOPERATIVE_GROUPS = True
COOPERATIVE_GROUP_SIZE = 4
USE_MURMUR_HASH = True
USE_AOS = True


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: " +
                ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(re.search(r"version\s*([\d.]+)",
                                         out.decode()).group(1))
            if cmake_version < "3.1.0":
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(
            os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ["-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + extdir,
                      "-DPYTHON_EXECUTABLE=" + sys.executable]

        cfg = "Debug" if self.debug else "Release"
        build_args = ["--config", cfg]

        if platform.system() == "Windows":
            cmake_args += ["-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(
                cfg.upper(),
                extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ["-A", "x64"]
            build_args += ["--", "/m"]
        else:
            cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]
            build_args += ["--", "-j2"]

        global CUDA_ERROR_CHECK 
        global USE_COOPERATIVE_GROUPS 
        global COOPERATIVE_GROUP_SIZE
        global USE_MURMUR_HASH
        cmake_args += ["-DCUDA_ERROR_CHECK=" + str(CUDA_ERROR_CHECK)]
        cmake_args += ["-DUSE_COOPERATIVE_GROUPS=" + str(USE_COOPERATIVE_GROUPS)]
        if COOPERATIVE_GROUP_SIZE in [1, 2, 4, 8, 16, 32]:
            cmake_args += ["-DCOOPERATIVE_GROUP_SIZE=" + str(COOPERATIVE_GROUP_SIZE)]
        cmake_args += ["-DUSE_MURMUR_HASH=" + str(USE_MURMUR_HASH)]

        env = os.environ.copy()
        env["CXXFLAGS"] = "{} -DVERSION_INFO=\\'{}\\'".format(
            env.get("CXXFLAGS", ""),
            self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(["cmake", ext.sourcedir] + cmake_args,
                              cwd=self.build_temp, env=env)
        subprocess.check_call(["cmake", "--build", "."] + build_args,
                              cwd=self.build_temp)


setup(
    name="kage-cucounter",
    version="0.0.2",
    author="Jørgen Henriksen",
    description="A hashtable based counter implemented in CUDA for high throughput.",
    long_description="",
    packages=["cucounter"],
    ext_modules=[CMakeExtension("cucounter_C")],
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
)

